# This file is a part of the CaosDB Project.
#
# Copyright (C) 2022 Indiscale GmbH <info@indiscale.com>
#               2022 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from pytest import fixture, mark

import caosdb as db
from caosdb.cached import cache_clear
from caosadvancedtools.models.parser import parse_model_from_string

from caoscrawler.crawl import Crawler
from caoscrawler.identifiable_adapters import CaosDBIdentifiableAdapter
from caoscrawler.structure_elements import DictElement

from caoscrawler.scanner import create_converter_registry, scan_structure_elements

from caosdb.utils.register_tests import clear_database, set_test_key
set_test_key("10b128cf8a1372f30aa3697466bb55e76974e0c16a599bb44ace88f19c8f61e2")


@fixture(autouse=True)
def clear_cache():
    cache_clear()


def test_issue_23(clear_database):
    """Test that an update leaves existing properties, that were not found by
    the crawler, unchanged.

    See issue https://gitlab.com/caosdb/caosdb-crawler/-/issues/23

    """

    # insert a simplistic model an arecord of type TestType with identifying
    # property and prop_a, but not prop_b.
    prop_ident = db.Property(name="identifying_prop", datatype=db.TEXT)
    prop_a = db.Property(name="prop_a", datatype=db.TEXT)
    prop_b = db.Property(name="prop_b", datatype=db.TEXT)
    rt = db.RecordType(name="TestType")
    rec = db.Record(name="TestRec").add_parent(rt)
    rec.add_property(name="identifying_prop", value="identifier")
    rec.add_property(name="prop_a", value="something")
    db.Container().extend([prop_ident, prop_a, prop_b, rt, rec]).insert()

    # set up crawler, first cfood defining a TestType record with
    # identifying_prop and prop_b, but not prop_a ...
    crawler_definition = {
        "DictTest": {
            "type": "DictElement",
            "match": "(.*)",
            "records": {
                "TestType": {}
            },
            "subtree": {
                "identifying_element": {
                    "type": "TextElement",
                    "match_name": "ident",
                    "match_value": "(?P<ident_value>.*)",
                    "records": {
                        "TestType": {
                            "identifying_prop": "$ident_value"
                        }
                    }
                },
                "other_element": {
                    "type": "TextElement",
                    "match_name": "prop_b",
                    "match_value": "(?P<other_value>.*)",
                    "records": {
                        "TestType": {
                            "prop_b": "$other_value"
                        }
                    }
                }
            }
        }
    }

    # register identifiable for TestType
    ident = CaosDBIdentifiableAdapter()
    ident.register_identifiable("TestType", db.RecordType().add_parent(
        name="TestType").add_property(name="identifying_prop"))

    crawler = Crawler(identifiableAdapter=ident)
    converter_registry = create_converter_registry(crawler_definition)

    # the dictionary to be crawled...
    test_dict = {
        "ident": "identifier",
        "prop_b": "something_else"
    }

    crawler.generate_run_id()
    records = scan_structure_elements(
        DictElement("TestDict", test_dict), crawler_definition, converter_registry)

    assert len(records) == 1
    rec_crawled = records[0]
    assert rec_crawled.parents[0].name == "TestType"
    assert rec_crawled.get_property("identifying_prop") is not None
    assert rec_crawled.get_property("identifying_prop").value == "identifier"
    assert rec_crawled.get_property("prop_b") is not None
    assert rec_crawled.get_property("prop_b").value == "something_else"
    # no interaction with the database yet, so the record shouldn't have a prop_a yet
    assert rec_crawled.get_property("prop_a") is None

    # synchronize with database and update the record
    ins, ups = crawler.synchronize(crawled_data=records)
    assert len(ins) == 0
    assert len(ups) == 1

    # retrieve and check that name and properties have been combined correctly
    rec_retrieved = db.Record(id=rec.id).retrieve()
    assert rec_retrieved.name == rec.name
    assert rec_retrieved.get_property(
        "identifying_prop").value == rec.get_property("identifying_prop").value
    assert rec_retrieved.get_property(
        "prop_a").value == rec.get_property("prop_a").value
    assert rec_retrieved.get_property(
        "identifying_prop").value == rec_crawled.get_property("identifying_prop").value
    assert rec_retrieved.get_property(
        "prop_b").value == rec_crawled.get_property("prop_b").value


def test_issue_83(clear_database):
    """https://gitlab.com/linkahead/linkahead-crawler/-/issues/83. Test that
    names don't need to be unique for referenced entities if they are not part
    of the identifiable.

    """

    # Very simple data model
    identifying_prop = db.Property(name="IdentifyingProp", datatype=db.INTEGER).insert()
    referenced_type = db.RecordType(name="ReferencedType").add_property(
        name=identifying_prop.name, importance=db.OBLIGATORY).insert()
    referencing_type = db.RecordType(name="ReferencingType").add_property(
        name=referenced_type.name, datatype=db.LIST(referenced_type.name)).insert()

    # Define identifiables. ReferencingType by name, ReferencedType by
    # IdentifyingProp and not by name.
    ident = CaosDBIdentifiableAdapter()
    ident.register_identifiable(referenced_type.name, db.RecordType().add_parent(
        name=referenced_type.name).add_property(name=identifying_prop.name))
    ident.register_identifiable(referencing_type.name, db.RecordType().add_parent(
        name=referencing_type.name).add_property(name="name"))

    crawler = Crawler(identifiableAdapter=ident)

    ref_target1 = db.Record(name="RefTarget").add_parent(
        name=referenced_type.name).add_property(name=identifying_prop.name, value=1)
    ref_target2 = db.Record(name="RefTarget").add_parent(
        name=referenced_type.name).add_property(name=identifying_prop.name, value=2)

    referencing1 = db.Record(name="Referencing1").add_parent(
        name=referencing_type.name).add_property(name=referenced_type.name, value=[ref_target1])
    referencing2 = db.Record(name="Referencing2").add_parent(
        name=referencing_type.name).add_property(name=referenced_type.name, value=[ref_target2])
    referencing3 = db.Record(name="Referencing3").add_parent(name=referencing_type.name).add_property(
        name=referenced_type.name, value=[ref_target1, ref_target2])

    records = db.Container().extend(
        [ref_target1, ref_target2, referencing1, referencing2, referencing3])

    ins, ups = crawler.synchronize(crawled_data=records, unique_names=False)
    assert len(ins) == len(records)
    assert len(ups) == 0

    retrieved_target1 = db.execute_query(
        f"FIND {referenced_type.name} WITH {identifying_prop.name}=1", unique=True)
    retrieved_target2 = db.execute_query(
        f"FIND {referenced_type.name} WITH {identifying_prop.name}=2", unique=True)
    assert retrieved_target2.name == retrieved_target1.name
    assert retrieved_target1.name == ref_target1.name
    assert retrieved_target1.id != retrieved_target2.id

    retrieved_referencing1 = db.execute_query(
        f"FIND {referencing_type.name} WITH name={referencing1.name}", unique=True)
    assert retrieved_referencing1.get_property(referenced_type.name) is not None
    assert retrieved_referencing1.get_property(referenced_type.name).value == [
        retrieved_target1.id]
    assert retrieved_referencing1.get_property(referenced_type.name).value != [
        retrieved_target2.id]

    retrieved_referencing2 = db.execute_query(
        f"FIND {referencing_type.name} WITH name={referencing2.name}", unique=True)
    assert retrieved_referencing2.get_property(referenced_type.name) is not None
    assert retrieved_referencing2.get_property(referenced_type.name).value == [
        retrieved_target2.id]
    assert retrieved_referencing2.get_property(referenced_type.name).value != [
        retrieved_target1.id]

    retrieved_referencing3 = db.execute_query(
        f"FIND {referencing_type.name} WITH name={referencing3.name}", unique=True)
    assert retrieved_referencing3.get_property(referenced_type.name) is not None
    assert len(retrieved_referencing3.get_property(referenced_type.name).value) == 2
    assert retrieved_target1.id in retrieved_referencing3.get_property(referenced_type.name).value
    assert retrieved_target2.id in retrieved_referencing3.get_property(referenced_type.name).value


def test_indiscale_113(clear_database):
    """Somewhat mysterious failures to resolve references in
    split_into_inserts_and_updates, see
    https://gitlab.indiscale.com/caosdb/src/caosdb-crawler/-/issues/113

    """

    # Create and insert minimal datamodel
    datamodel_str = """
Dataset:
  recommended_properties:
    Title:
      datatype: TEXT
    Abstract:
      datatype: TEXT
    Event:
Published Dataset:
  inherit_from_suggested:
    - Dataset
  obligatory_properties:
    DOI:
      datatype: TEXT
Event:
  obligatory_properties:
    Start datetime:
      datatype: DATETIME
    longitude:
      datatype: DOUBLE
    latitude:
      datatype: DOUBLE
  recommended_properties:
    label:
      datatype: TEXT
    Basis:
    Campaign:
Basis:
Campaign:
  recommended_properties:
    Basis:
"""
    model = parse_model_from_string(datamodel_str)
    model.sync_data_model(noquestion=True)

    # Register identifiables for Dataset (title), Published Dataset (DOI), Event
    # (Start datetime, latitude, longitude, label), Basis (name), and Campaign
    # (name).
    ident = CaosDBIdentifiableAdapter()
    ident.register_identifiable("Dataset", db.RecordType().add_parent(
        name="Dataset").add_property(name="title"))
    ident.register_identifiable("Published Dataset", db.RecordType(
    ).add_parent(name="Published Dataset").add_property(name="DOI"))
    ident.register_identifiable("Event", db.RecordType().add_parent(name="Event").add_property(name="longitude").add_property(
        name="latitude").add_property(name="Start datetime").add_property(name="label"))
    ident.register_identifiable("Basis", db.RecordType().add_parent(
        name="Basis").add_property(name="name"))
    ident.register_identifiable("Campaign", db.RecordType().add_parent(
        name="Campaign").add_property(name="name"))

    crawler = Crawler(identifiableAdapter=ident)

    xmls = ["""<Record>
  <Parent name="Published Dataset"/>
  <Property name="Title" importance="FIX" flag="inheritance:FIX">Chemical analysis from sediment cores in the area of a landslide scar close to the airport of Nice</Property>
  <Property name="Abstract" importance="FIX" flag="inheritance:FIX">Submarine groundwater discharge (SGD), the flow of fresh and saline groundwater from the seabed into the coastal ocean, has been intensively investigated in the recent years. This research has usually been restricted to shallow water and intertidal areas, whereas knowledge about groundwater seepage in deeper water is mainly limited to point sources from karstic aquifers. In this study we observed submarine groundwater seepage and a subterranean estuary in sediments at water depths of 20-44 m located within the Ligurian Margin, western Mediterranean Sea. Here, a catastrophic submarine landslide occurred near the Nice airport (French Ligurian coast) in the fall of 1979 after a period of heavy rainfall. During two research cruises, gravity cores were recovered in and around the area of the landslide scar. Pore water samples collected from sediment cores indicated sediments containing freshwater within the landslide scar. Pore water profiles of selected ions, such as chloride, ammonium, manganese, sulfate and barium were used to assess transport and reaction processes within the sediment. A 1-dimensional transport model indicates in most cores upward pore water velocities of 2.3-8.8 cm/yr. This study shows that submarine groundwater seepage along the French Mediterranean coastline can occur at water depths reaching 44 m.</Property>
  <Property name="DOI" importance="FIX" flag="inheritance:FIX">https://doi.org/10.1594/PANGAEA.882187</Property>
  <Property name="Dataset" datatype="LIST&lt;Dataset&gt;" importance="FIX" flag="inheritance:FIX">
    <Value>&lt;Record&gt;
  &lt;Parent name="Published Dataset"/&gt;
  &lt;Property name="Title" importance="FIX" flag="inheritance:FIX"&gt;(Appendix 3) Different saturation index analysis from sediment cores in the area of a landslide scar close to the airport of Nice&lt;/Property&gt;
  &lt;Property name="Abstract" importance="FIX" flag="inheritance:FIX"/&gt;
  &lt;Property name="DOI" importance="FIX" flag="inheritance:FIX"&gt;https://doi.org/10.1594/PANGAEA.882184&lt;/Property&gt;
  &lt;Property name="Event" datatype="LIST&amp;lt;Event&amp;gt;" importance="FIX" flag="inheritance:FIX"&gt;
    &lt;Value&gt;GeoB13952&lt;/Value&gt;
  &lt;/Property&gt;
&lt;/Record&gt;
</Value>
  </Property>
</Record>""", """<Record>
  <Parent name="Published Dataset"/>
  <Property name="Title" importance="FIX" flag="inheritance:FIX">(Appendix 3) Different saturation index analysis from sediment cores in the area of a landslide scar close to the airport of Nice</Property>
  <Property name="Abstract" importance="FIX" flag="inheritance:FIX"/>
  <Property name="DOI" importance="FIX" flag="inheritance:FIX">https://doi.org/10.1594/PANGAEA.882184</Property>
  <Property name="Event" datatype="LIST&lt;Event&gt;" importance="FIX" flag="inheritance:FIX">
    <Value>GeoB13952</Value>
  </Property>
</Record>""", """<Record name="GeoB13952">
  <Parent name="Event"/>
  <Property name="label" importance="FIX" flag="inheritance:FIX">GeoB13952</Property>
  <Property name="Start datetime" importance="FIX" flag="inheritance:FIX">2009-07-03T07:52:36</Property>
  <Property name="longitude" importance="FIX" flag="inheritance:FIX">7.22225</Property>
  <Property name="latitude" importance="FIX" flag="inheritance:FIX">43.6438</Property>
  <Property name="Basis" importance="FIX" flag="inheritance:FIX">Poseidon</Property>
  <Property name="Campaign" importance="FIX" flag="inheritance:FIX">POS386</Property>
</Record>""", """<Record name="GeoB13952">
  <Parent name="Event"/>
  <Property name="label" importance="FIX" flag="inheritance:FIX">GeoB13952</Property>
  <Property name="Start datetime" importance="FIX" flag="inheritance:FIX">2009-07-03T07:52:36</Property>
  <Property name="longitude" importance="FIX" flag="inheritance:FIX">7.22225</Property>
  <Property name="latitude" importance="FIX" flag="inheritance:FIX">43.6438</Property>
  <Property name="Basis" importance="FIX" flag="inheritance:FIX">Poseidon</Property>
  <Property name="Campaign" importance="FIX" flag="inheritance:FIX">POS386</Property>
</Record>""", """<Record>
  <Parent name="Published Dataset"/>
  <Property name="Title" importance="FIX" flag="inheritance:FIX">(Appendix 3) Different saturation index analysis from sediment cores in the area of a landslide scar close to the airport of Nice</Property>
  <Property name="Abstract" importance="FIX" flag="inheritance:FIX"/>
  <Property name="DOI" importance="FIX" flag="inheritance:FIX">https://doi.org/10.1594/PANGAEA.882184</Property>
  <Property name="Event" datatype="LIST&lt;Event&gt;" importance="FIX" flag="inheritance:FIX">
    <Value>GeoB13952</Value>
  </Property>
</Record>""", """<Record name="GeoB13952">
  <Parent name="Event"/>
  <Property name="label" importance="FIX" flag="inheritance:FIX">GeoB13952</Property>
  <Property name="Start datetime" importance="FIX" flag="inheritance:FIX">2009-07-03T07:52:36</Property>
  <Property name="longitude" importance="FIX" flag="inheritance:FIX">7.22225</Property>
  <Property name="latitude" importance="FIX" flag="inheritance:FIX">43.6438</Property>
  <Property name="Basis" importance="FIX" flag="inheritance:FIX">Poseidon</Property>
  <Property name="Campaign" importance="FIX" flag="inheritance:FIX">POS386</Property>
</Record>""", """<Record name="Poseidon">
  <Parent name="Basis"/>
</Record>""", """<Record name="POS386">
  <Parent name="Campaign"/>
  <Property name="Basis" importance="FIX" flag="inheritance:FIX">Poseidon</Property>
</Record>"""
            ]
    # recs = db.Container().extend([db.common.models.parse_xml(xml_rec) for xml_rec in xmls])
    basis = db.Record(name="Poseidon").add_parent(name="Basis")
    campaign = db.Record(name="POS386").add_parent(name="Campaign").add_property(name="Basis", value=basis)
    event = db.Record(name="GeoB13952").add_parent(name="Event").add_property(name="label", value="GeoB13952")
    event.add_property(name="Start datetime", value="2009-07-03T07:52:36")
    event.add_property(name="longitude", value=7.22225)
    event.add_property(name="latitude", value=43.6438)
    event.add_property(name="Basis", value=basis)
    event.add_property(name="Campaign", value=campaign)
    ds_child = db.Record().add_parent(name="Published Dataset")
    ds_child.add_property(name="Title", value="(Appendix 3) Different saturation index analysis from sediment cores in the area of a landslide scar close to the airport of Nice")
    ds_child.add_property(name="Abstract")
    ds_child.add_property(name="DOI", value="https://doi.org/10.1594/PANGAEA.882184")
    ds_child.add_property(name="Event", value=[event], datatype=db.LIST("Event"))
    ds_parent = db.Record().add_parent(name="Published Dataset")
    ds_parent.add_property(name="Title", value="Chemical analysis from sediment cores in the area of a landslide scar close to the airport of Nice")
    ds_parent.add_property(name="Abstract", value="Submarine groundwater discharge (SGD), the flow of fresh and saline groundwater from the seabed into the coastal ocean, has been intensively investigated in the recent years. This research has usually been restricted to shallow water and intertidal areas, whereas knowledge about groundwater seepage in deeper water is mainly limited to point sources from karstic aquifers. In this study we observed submarine groundwater seepage and a subterranean estuary in sediments at water depths of 20-44 m located within the Ligurian Margin, western Mediterranean Sea. Here, a catastrophic submarine landslide occurred near the Nice airport (French Ligurian coast) in the fall of 1979 after a period of heavy rainfall. During two research cruises, gravity cores were recovered in and around the area of the landslide scar. Pore water samples collected from sediment cores indicated sediments containing freshwater within the landslide scar. Pore water profiles of selected ions, such as chloride, ammonium, manganese, sulfate and barium were used to assess transport and reaction processes within the sediment. A 1-dimensional transport model indicates in most cores upward pore water velocities of 2.3-8.8 cm/yr. This study shows that submarine groundwater seepage along the French Mediterranean coastline can occur at water depths reaching 44 m.")
    ds_parent.add_property(name="DOI", value="https://doi.org/10.1594/PANGAEA.882187")
    ds_parent.add_property(name="Dataset", datatype=db.LIST("Dataset"), value=ds_child)

    # multiples are necessary
    recs = [ds_parent, ds_child, event, event, ds_child, event, basis, campaign]
    
    ins, ups = crawler.synchronize(crawled_data=recs, unique_names=False)
